#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .value import MBIOValue

from .task import MBIOTask
from .xmlconfig import XMLConfig

import threading
import base64
import httpx
import pprint
import json

from prettytable import PrettyTable

# https://developers.meethue.com/develop/get-started-2/#follow-3-easy-steps
# httdevicesps://developers.meethue.com/develop/hue-api-v2/api-reference/


class HueResources(object):
    RESSOURCE_NAME=''

    def __init__(self, hue: Hue):
        self._hue=hue
        self._items={}
        self._services={}
        self.update()

    @property
    def hue(self):
        return self._hue

    @property
    def logger(self):
        return self.hue.logger

    def items(self):
        return self._items.values()

    def services(self):
        return self._services.values()

    def update(self):
        try:
            r=self.hue.GET(self.hue.urlclip('resource/%s' % self.RESSOURCE_NAME))
            if r:
                data=r['data']
                for item in data:
                    self._items[item['id']]=item
                self.rebuildServicesIndex()
                return data
        except:
            pass

    def get(self, rid):
        if not self._items:
            self.update()
        try:
            return self._items[rid]
        except:
            pass

    def search(self, name):
        items=[]
        if name:
            name=name.strip().lower()
            for item in self._items.values():
                try:
                    if name in item['metadata']['name'].lower():
                        items.append(item['id'])
                except:
                    pass
        return items

    def first(self, name):
        try:
            return self.search(name)[0]
        except:
            pass

    def dumpdata(self):
        pprint.pprint(self._items)

    def dump(self):
        t=PrettyTable()
        t.field_names=['TYPE', 'id', '...', 'name']
        t.align='l'

        for item in self.items():
            name=item['metadata']['name'].lower()
            t.add_row([self.__class__.__name__, item['id'], item['id_v1'], name])

        for service in self.services():
            name=service['itemrid']
            item=self.get(service['itemrid'])
            if item:
                name=item['metadata']['name'].lower()
            t.add_row(['service', service['rid'], service['rtype'], '->%s' % name])

        print(t.get_string())

    def rebuildServicesIndex(self, force=False):
        if force or not self._services:
            for item in self.items():
                try:
                    for service in item['services']:
                        rtype=service['rtype']
                        rid=service['rid']
                        itemid=item['id']
                        self._services[rid]={'rtype': rtype, 'rid': rid, 'itemrid': itemid}
                except:
                    self.logger.exception('rebuildIndex')
                    pass

    # FIXME: not used
    def getItemFromServiceId(self, rid) -> dict:
        try:
            service=self._services[rid]
            if service:
                return self.get(service['itemrid'])
        except:
            pass

    def getService(self, rtype, itemrid) -> str:
        item=self.get(itemrid)
        if item and rtype:
            try:
                # TODO: loop over small entries, but may be better
                for service in item['services']:
                    if service['rtype']==rtype:
                        return service['rid']
            except:
                pass


class HueService(object):
    def __init__(self, hue: Hue, rtype, rid):
        self._hue=hue
        self._rtype=rtype
        self._rid=rid
        self._data={}

    @property
    def hue(self):
        return self._hue

    @property
    def rid(self):
        return self._rid

    @property
    def rtype(self):
        return self._rtype

    def url(self):
        return self.hue.urlclip('resource/%s/%s' % (self.rtype, self.rid))

    @property
    def data(self):
        if not self._data:
            self.update()
        return self._data

    def update(self):
        try:
            url=self.hue.urlclip('resource/%s/%s' % (self.rtype, self.rid))
            r=self.hue.GET(url)
            if r:
                self._data=r['data'][0]
                return self._data
        except:
            pass


class HueServiceLight(HueService):
    def getStateFromData(self, data):
        try:
            return bool(data['on']['on'])
        except:
            pass

    def getLevelFromData(self, data):
        try:
            return float(data['dimming']['brightness'])
        except:
            pass

    def state(self):
        return self.getStateFromData(self.data)

    def level(self):
        return self.getLevelFromData(self.data)

    def setState(self, state=True):
        try:
            data={'on': {'on': bool(state)}}
            r=self.hue.PUT(self.url(), data=data)
            if r is not None:
                return True
        except:
            pass

    def on(self):
        return self.setState(True)

    def off(self):
        return self.setState(False)

    def toggle(self):
        self.setState(not self.state())

    def setLevel(self, level):
        try:
            data={'dimming': {'brightness': float(level)}}
            r=self.hue.PUT(self.url(), data=data)
            if r is not None:
                return True
        except:
            pass

    def max(self):
        return self.setLevel(100.0)


class HueDevices(HueResources):
    RESSOURCE_NAME='device'

    def light(self, rid) -> HueServiceLight:
        rtype='light'
        servicerid=self.getService(rtype, rid)
        if servicerid:
            return HueServiceLight(self.hue, rtype, servicerid)


class HueRooms(HueResources):
    RESSOURCE_NAME='room'

    def light(self, rid) -> HueServiceLight:
        rtype='grouped_light'
        servicerid=self.getService(rtype, rid)
        if servicerid:
            return HueServiceLight(self.hue, rtype, servicerid)


class HueZones(HueResources):
    RESSOURCE_NAME='zone'

    def light(self, rid) -> HueServiceLight:
        rtype='grouped_light'
        servicerid=self.getService(rtype, rid)
        if servicerid:
            return HueServiceLight(self.hue, rtype, servicerid)


class Hue(object):
    def __init__(self, bridge, token, logger):
        self._logger=logger
        self._bridge=bridge
        self._token=token
        self._devices=HueDevices(self)
        self._rooms=HueRooms(self)
        self._zones=HueZones(self)

    @property
    def logger(self):
        return self._logger

    @property
    def devices(self):
        return self._devices

    @property
    def rooms(self):
        return self._rooms

    @property
    def zones(self):
        return self._zones

    def url(self, rootpath, subpath=None):
        url='https://%s/%s' % (self._bridge, rootpath)
        if subpath:
            return '%s/%s' % (url, subpath)
        return url

    def urlclip(self, subpath=None):
        return self.url('clip/v2', subpath)

    def urlapi(self, subpath=None):
        return self.url('api', subpath)

    def isAuth(self):
        if self._token:
            return True

    def GET(self, url, data={}):
        try:
            self.logger.debug('GET %s(%s)' % (url, data))
            headers={}
            if self.isAuth():
                headers['hue-application-key']=self._token
            r=httpx.get(url, headers=headers, params=data, verify=False, timeout=3.0)
            if r and r.status_code==httpx.codes.OK:
                data=r.json()
                # self.logger.debug(data)
                return data
            self.logger.error(r.text)
        except:
            # self.logger.exception('GET')
            pass

    def POST(self, url, data={}):
        try:
            self.logger.debug('POST %s(%s)' % (url, data))
            headers={}
            if self.isAuth():
                headers['hue-application-key']=self._token
            r=httpx.post(url, headers=headers, json=data, verify=False, timeout=3.0)
            if r and r.status_code==httpx.codes.OK:
                data=r.json()
                # self.logger.debug(data)
                return data
            self.logger.error(r.text)
        except:
            # self.logger.exception('POST')
            pass

    def PUT(self, url, data={}):
        try:
            self.logger.debug('PUT %s(%s)' % (url, data))
            headers={}
            if self.isAuth():
                headers['hue-application-key']=self._token
            r=httpx.put(url, headers=headers, json=data, verify=False, timeout=3.0)
            if r and r.status_code==httpx.codes.OK:
                data=r.json()
                self.logger.debug(data)
                return data
            self.logger.error(r.text)
        except:
            # self.logger.exception('PUT')
            pass

    def events(self):
        url=self.url('eventstream', 'clip/v2')
        headers={'Accept': 'text/event-stream'}
        if self.isAuth():
            headers['hue-application-key']=self._token
            try:
                with httpx.stream('GET', url, headers=headers, verify=False, timeout=60) as s:
                    # yield from s.iter_bytes()
                    yield from s.iter_lines()
            except:
                pass

    def ping(self):
        r=self.GET(self.urlapi('0/config'))
        return r

    def auth(self, clientid=None):
        if not self.isAuth():
            if not clientid:
                clientid='MBIOPROCESSOR'
            try:
                url=self.urlapi()
                data={'devicetype': clientid, 'generateclientkey': True}
                r=self.POST(url, data=data)
                data=r[0]['success']
                if data:
                    self._token=data['username']
            except:
                self.logger.exception('auth')
        try:
            return '(b16)%s' % base64.b16encode(self._token.encode()).decode()
        except:
            pass

    def discover(self):
        bridges={}

        r=httpx.get('https://discovery.meethue.com', verify=False, timeout=3.0)
        if r and r.status_code==httpx.codes.OK:
            data=r.json()
            try:
                for bridge in data:
                    bridges[bridge['id'].lower()]=bridge
            except:
                pass

        return bridges

    def getBridgeIp(self, bid):
        bridges=self.discover()
        try:
            return bridges[bid.lower()]['internalipaddress']
        except:
            pass


class HueResourceBag(object):
    # Used to store local info/ressources
    # Not a god name but nothing best found ;(
    # Proprietary format, not Hue format

    def __init__(self, resources: HueResources):
        self._resources=resources
        self._data={}
        self._values=[]

    @property
    def hue(self):
        return self._resources.hue

    @property
    def logger(self):
        return self.hue.logger

    def get(self, rid) -> dict:
        try:
            return self._data[rid]
        except:
            pass

    def items(self):
        return self._data.values()

    def values(self):
        return self._values

    def declare(self, rid, name, values):
        if not self.get(rid) and name:
            nonemptyvalues={}
            for value in values:
                if values[value] is not None:
                    nonemptyvalues[value]=values[value]

            data={'rid': rid, 'name': name, 'values': nonemptyvalues}
            self._data[rid]=data
            if values:
                for value in values.values():
                    self._values.append(value)
            return data

    def value(self, rid, name) -> MBIOValue:
        try:
            data=self.get(rid)
            return data['values'][name]
        except:
            pass

    def dump(self):
        t=PrettyTable()
        t.field_names=['Type', 'rid', 'Name']
        t.align='l'

        for item in self.items():
            t.add_row([self.__class__.__name__, item['rid'], item['name']])

        print(t.get_string())

    def updateValue(self, resource, name, v):
        try:
            if resource and resource['values']:
                try:
                    resource['values'][name].updateValue(v)
                except:
                    pass
        except:
            pass


class MBIOTaskHue(MBIOTask):
    def initName(self):
        return 'hue'

    @property
    def hue(self):
        return self._hue

    @property
    def rooms(self):
        return self._rooms

    @property
    def zones(self):
        return self._zones

    def onInit(self):
        self._hue=None
        self._rooms=None
        self._zones=None
        self.config.set('refreshperiod', 5)
        self._timeoutRefresh=0
        self._retry=3
        self.valueDigital('comerr', default=False)
        self._threadHueEvents=None

    def onLoad(self, xml: XMLConfig):
        self.config.set('bridge', xml.get('bridge'))
        self.config.set('token', xml.get('token'))
        self.config.update('refreshperiod', xml.getInt('refresh'))
        self._hue=Hue(self.config.bridge, self.config.token, self.logger)
        self._rooms=HueResourceBag(self.hue.rooms)
        self._zones=HueResourceBag(self.hue.zones)

        for room in xml.children('room'):
            name=room.get('name')
            rid=room.get('id')
            dimmable=room.getBool('dimmable')
            if name:
                state=self.valueDigital('%s_state' % name, writable=True)
                if dimmable:
                    level=self.value('%s_level' % name, unit='%', resolution=0.1, writable=True)
                    level.setRange(0, 100)
                else:
                    level=None
                self.rooms.declare(rid, name, values={'state': state, 'level': level})

        for zone in xml.children('zone'):
            name=zone.get('name')
            rid=zone.get('id')
            dimmable=zone.getBool('dimmable')
            if name:
                state=self.valueDigital('%s_state' % name, writable=True)
                if dimmable:
                    level=self.value('%s_level' % name, unit='%', resolution=0.1, writable=True)
                    level.setRange(0, 100)
                else:
                    level=None
                self.zones.declare(rid, name, values={'state': state, 'level': level})

    def hueEventManager(self):
        halt=False

        self.logger.info('Starting starting hue events thread')
        while not halt:
            for event in self.hue.events():
                if self.state not in [self.STATE_RUN]:
                    halt=1
                    break

                if not halt and event[:6]=='data: ':
                    try:
                        # self.logger.warning(event)
                        data=json.loads(event[6:])
                        for e in data:
                            # eid=data['id']
                            etype=e['type']
                            edata=e['data']
                            for item in edata:
                                self.dispatchEvent(etype, item)
                    except:
                        self.logger.exception('event')

            if not halt:
                # Auto restart event stream if closed
                self.sleep(5)

        self.logger.info('Exiting hue events thread')
        self._threadHueEvents=None

    def dispatchEvent(self, etype, data):
        self.logger.debug('HUE/EVENT %s %s' % (etype, data))
        if etype=='update':
            try:
                owner=data['owner']['rid']
            except:
                return

            room=self.rooms.get(owner)
            if room:
                service=self.hue.rooms.light(room['rid'])

                v=service.getStateFromData(data)
                self.rooms.updateValue(room, 'state', v)
                v=service.getLevelFromData(data)
                self.rooms.updateValue(room, 'level', v)
                return

            zone=self.zones.get(owner)
            if zone:
                service=self.hue.zones.light(zone['rid'])

                v=service.getStateFromData(data)
                self.zones.updateValue(room, 'state', v)
                v=service.getLevelFromData(data)
                self.zones.updateValue(room, 'level', v)
                return

    def poweron(self):
        ip=self.hue.getBridgeIp(self.config.bridge)
        if ip:
            self.logger.info('Bridge %s found @%s' % (self.config.bridge, ip))
            self.config.bridge=ip

        if self._hue.ping():
            self.hue.devices.update()
            self.hue.rooms.update()
            self.hue.zones.update()
            return True

    def poweroff(self):
        return True

    def syncRooms(self):
        for room in self.rooms.items():
            self.microsleep()
            values=room['values'].values()
            service=self.hue.rooms.light(room['rid'])
            if values and service:
                for value in values:
                    if value.isPendingSync():
                        if service.setState(value.toReachValue):
                            value.clearSyncAndUpdateValue()

    def syncZones(self):
        for zone in self.zones.items():
            self.microsleep()
            values=zone['values'].values()
            service=self.hue.zones.light(zone['rid'])
            if values and service:
                for value in values:
                    if value.isPendingSync():
                        if service.setState(value.toReachValue):
                            value.clearSyncAndUpdateValue()

    def refreshRooms(self):
        result=True
        for room in self.rooms.items():
            self.microsleep()

            service=self.hue.rooms.light(room['rid'])
            if service:
                # refresh data
                data=service.update()

                v=service.getStateFromData(data)
                self.rooms.updateValue(room, 'state', v)
                v=service.getLevelFromData(data)
                self.rooms.updateValue(room, 'level', v)
                continue

            for value in room['values'].values():
                if value:
                    value.setError(True)
            result=False

        return result

    def refreshZones(self):
        result=True
        for zone in self.zones.items():
            self.microsleep()

            service=self.hue.zones.light(zone['rid'])
            if service:
                # refresh data
                data=service.update()

                v=service.getStateFromData(data)
                self.rooms.updateValue(zone, 'state', v)
                v=service.getLevelFromData(data)
                self.rooms.updateValue(zone, 'level', v)
                continue

            for value in zone['values'].values():
                if value:
                    value.setError(True)
            result=False

        return result

    def run(self):
        if not self.hue.isAuth():
            for value in self.rooms.values():
                value.setError(True)
            for value in self.zones.values():
                value.setError(True)

            token=self.hue.auth()
            if token:
                # TODO: permanent storage
                self.hue._token=token
                return 1.0

            self.values.comerr.updateValue(True)
            return 15.0

        if not self._threadHueEvents:
            self._threadHueEvents=threading.Thread(target=self.hueEventManager, daemon=True)
            self._threadHueEvents.start()

        self.syncRooms()
        self.syncZones()

        if self.isTimeout(self._timeoutRefresh):
            self._timeoutRefresh=self.timeout(self.config.refreshperiod)
            error=False

            if not self.refreshRooms():
                error=True
            if not self.refreshZones():
                error=True

            if not error:
                self._timeoutRefresh=self.timeout(self.config.refreshperiod)
                self._retry=3
                self.values.comerr.updateValue(False)
            else:
                self._timeoutRefresh=self.timeout(60)
                if self._retry>0:
                    self._retry-=1
                    if self._retry==0:
                        self.values.comerr.updateValue(True)

        return 5.0


if __name__ == "__main__":
    pass



#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .mbio import MBIO

from .items import Items
# from .config import MBIOConfig
from .xmlconfig import XMLConfig
from .device import MBIODevices, MBIODeviceGeneric
from .value import MBIOValues, MBIOValueDigital


from .socket import MBIOSocketString

import time
import threading

from prettytable import PrettyTable

from pymodbus.client import ModbusTcpClient
import logging
import ipcalc


# Modbus Clients Examples
# https://pymodbus.readthedocs.io/en/dev/source/examples.html


class MBIOGateway(object):
    """MBIOGateway object, containing MBIOValues containing every MBIOValue"""
    def __init__(self, parent: MBIO, name, host, port=502, interface=None, timeout=3, retries=3, xml: XMLConfig = None):
        self._parent: MBIO = parent
        self._name=str(name).lower()
        if not name:
            name='gw%d' % parent.gateways.count()
        self._host=host
        self._model=None
        self._MAC=None
        self._password=None
        self._interface=interface
        self._port=port
        self._timeout=timeout
        self._retries=retries
        self._error=False

        self._key='%s' % self._name

        self._client: ModbusTcpClient=None
        self._timeoutClient=0
        self._timeoutIdleAfterSend=0

        self._devices=MBIODevices(parent.logger)
        self._sysvalues=MBIOValues(self, self.key, self.logger)
        self._sysComErr=MBIOValueDigital(self._sysvalues, 'comerr')

        self.logger.info('Declaring GW:%s', self.host)

        self._eventStop=threading.Event()

        self.load(xml)

        # FIXME: daemon=True ?
        self._thread=threading.Thread(target=self.manager)
        self.logger.info("Starting background GW:%s task" % self.host)
        self._thread.start()

    @property
    def parent(self) -> MBIO:
        return self._parent

    def getMBIO(self) -> MBIO:
        return self._parent

    @property
    def logger(self):
        return self._parent.logger

    def configurator(rself):
        # To be overriden
        pass

    def onLoad(self, xml: XMLConfig):
        pass

    def load(self, xml: XMLConfig):
        if xml:
            try:
                self._MAC=xml.get('mac')
                self._model=xml.get('model')
                self._password=xml.get('password')
                if xml.isConfig('gateway'):
                    # TODO: clone other gw feature
                    self.onLoad(xml)
            except:
                self.logger.exception('%s:%s:load()' % (self.__class__.__name__, self.key))

    @property
    def key(self):
        return self._key

    def timeout(self, delay):
        return time.time()+delay

    def isTimeout(self, t):
        if time.time()>=t:
            return True
        return False

    def isOpen(self):
        if self._client and self._client.is_socket_open():
            return True
        return False

    def open(self):
        if not self._client:
            try:
                self.logger.info('Connecting to GW %s:%d (interface %s)' % (self._host, self._port, self._interface))
                source=(self._interface, 0)
                if not self._interface:
                    source=None
                self._client=ModbusTcpClient(host=self._host, port=self._port,
                            source_address=source,
                            timeout=self._timeout,
                            retries=self._retries,
                            # broadcast_enable=True,
                            # strict=True,
                            close_comm_on_error=False)
                # FIXME: how to disable correctly pymodbus logging (connect errors) ?
                logger=logging.getLogger()
                if logger:
                    logger.setLevel(logging.CRITICAL)
                self._timeoutIdleAfterSend=0
            except:
                pass

        if self._client:
            if self._client.connected:
                return self._client
            try:
                if self.isTimeout(self._timeoutClient):
                    self._timeoutClient=self.timeout(15)
                    self.logger.info('Opening socket')
                    if self._client.connect():
                        return self._client
            except:
                pass

    def close(self):
        try:
            if self._client:
                self.logger.info('Closing socket')
                self._client.close()
        except:
            pass
        self._client=None

    @property
    def client(self):
        return self.open()

    @property
    def name(self):
        return self._name

    @property
    def host(self):
        return self._host

    @property
    def MAC(self):
        return self._MAC

    @property
    def model(self):
        return self._model

    @property
    def password(self):
        return self._password

    def probe(self, address):
        try:
            self.logger.debug('Probing device address %d' % address)
            self.checkIdleAfterSend()
            r=self.client.read_device_information(slave=address)
            self.signalMessageTransmission()
            if r and not r.isError():
                data={'address': address,
                      'vendor': r.information[0].decode(),
                      'model': r.information[1].decode(),
                      'version': r.information[2].decode()}
                self.logger.info('Found device [%s] [%s] %s at address %d' %
                                 (data['vendor'], data['model'], data['version'], address))
                return data
        except:
            pass

    # FIXME: implement a better MBIODevice class registration
    def declareDeviceFromName(self, vendor, model, address, xml: XMLConfig = None):
        try:
            address=int(address)
            model=model or 'unknown'

            if vendor and address>0:
                vendor=vendor.lower()
                model=model.lower()
                device=None

                # TODO: better design

                if 'metz' in vendor:
                    if 'di10' in model:
                        from .metzconnect import MBIODeviceMetzConnectMRDI10
                        device=MBIODeviceMetzConnectMRDI10(self, address, xml=xml)
                    if 'di4' in model:
                        from .metzconnect import MBIODeviceMetzConnectMRDI4
                        device=MBIODeviceMetzConnectMRDI4(self, address, xml=xml)
                    elif 'do4' in model:
                        from .metzconnect import MBIODeviceMetzConnectMRDO4
                        device=MBIODeviceMetzConnectMRDO4(self, address, xml=xml)
                    elif 'aop4' in model or 'ao4' in model:
                        from .metzconnect import MBIODeviceMetzConnectMRAOP4
                        device=MBIODeviceMetzConnectMRAOP4(self, address, xml=xml)
                    elif 'ai8' in model:
                        from .metzconnect import MBIODeviceMetzConnectMRAI8
                        device=MBIODeviceMetzConnectMRAI8(self, address, xml=xml)

                elif 'belimo' in vendor:
                    if 'p22rth' in model:
                        from .belimo import MBIODeviceBelimoP22RTH
                        device=MBIODeviceBelimoP22RTH(self, address, xml=xml)
                    if 'actuator' in model:
                        from .belimo import MBIODeviceBelimoActuator
                        device=MBIODeviceBelimoActuator(self, address, xml=xml)

                elif 'digimat' in vendor:
                    if 'plc' in model:
                        from .digimatplc import MBIODeviceDigimatPLC
                        device=MBIODeviceDigimatPLC(self, address, xml=xml)

                elif 'ebm' in vendor:
                    if 'base' in model:
                        from .ebm import MBIODeviceEBM
                        device=MBIODeviceEBM(self, address, xml=xml)

                elif 'generic' in vendor:
                    device=MBIODeviceGeneric(self, address, xml=xml)

                return device
        except:
            pass

    def discover(self, start=1, end=32, maxerrors=3):
        devices=[]
        errors=0
        for address in range(start, end+1):
            data=self.probe(address)
            if data:
                if not self.device(address):
                    device=self.declareDeviceFromName(data['vendor'], data['model'], address)
                    if device:
                        devices.append(device)
                continue
            errors+=1
            maxerrors-=1
            if maxerrors>0 and errors>maxerrors:
                break
        return devices

    def ping(self, address):
        try:
            self.checkIdleAfterSend()
            r=self.client.diag_read_bus_message_count(slave=address)
            self.signalMessageTransmission()
            if r:
                if not r.isError():
                    return True
            self.logger.error('Unable to ping device %d' % address)
        except:
            # self.logger.exception('ping')
            pass
        return False

    def stop(self):
        self.halt()
        self._eventStop.set()

    def waitForThreadTermination(self):
        self.stop()
        self._thread.join()
        self.close()

    def sleep(self, delay=1):
        try:
            if self._eventStop.is_set():
                return True
            return self._eventStop.wait(delay)
        except:
            pass

    def checkIdleAfterSend(self):
        while not self.isTimeout(self._timeoutIdleAfterSend):
            self.sleep(0.001)

    def signalMessageTransmission(self):
        # https://minimalmodbus.readthedocs.io/en/stable/serialcommunication.html
        # 19200bps->2ms
        self._timeoutIdleAfterSend=self.timeout(0.002)

    def manager(self):
        while True:
            self.sleep(0.1)
            try:
                halted=True
                error=False
                for device in self._devices:
                    device.manager()
                    device.microsleep()
                    if device.isError():
                        error=True
                    if not device.isHalted():
                        halted=False

                self._error=error
                self._sysComErr.updateValue(self._error)

                if self._eventStop.is_set():
                    if halted:
                        self.logger.info("Exiting background GW:%s task" % self.host)
                        self.close()
                        return
            except:
                self.logger.exception("Background GW:%s task" % self.host)

    def isError(self):
        return self._error

    @property
    def devices(self):
        return self._devices

    def device(self, did):
        return self._devices.item(did)

    def reset(self, address=None):
        if address is not None:
            try:
                self.device(address).reset()
            except:
                pass
        else:
            self._devices.reset()

    def resetHalted(self):
        self._devices.resetHalted()

    def halt(self, address=None):
        if address is not None:
            try:
                self.device(address).halt()
            except:
                pass
        else:
            self._devices.halt()

    def dump(self):
        if self._devices:
            t=PrettyTable()
            t.field_names=['ADR', 'Key', 'Vendor', 'Model', 'Version', 'Class', 'State', 'Error', 'Values']
            t.align='l'
            for device in self._devices:
                t.add_row([device.address, device.key, device.vendor, device.model, device.version,
                           device.__class__.__name__, device.statestr(), str(device.isError()), device.values.count()])

        print(t.get_string(sortby="ADR"))

    def count(self):
        return len(self._devices)

    def __getitem__(self, key):
        return self.device(key)


    def __repr__(self):
        state='CLOSED'
        if self.isOpen():
            state='OPEN'
        return '%s(%s=%s, %d devices, %s)' % (self.__class__.__name__, self.name, self.host, self.count(), state)

    # FIXME: debug
    def toggle(self):
        for device in self.devices:
            try:
                device.toggle()
            except:
                pass

    def on(self):
        for device in self.devices:
            try:
                device.on()
            except:
                pass

    def off(self):
        for device in self.devices:
            try:
                device.off()
            except:
                pass

    def auto(self):
        for device in self.devices:
            device.auto()

    def manual(self):
        for device in self.devices:
            device.manual()


class MBIOGatewayMetzConnect(MBIOGateway):
    def configurator(self):
        return MBIOGatewayMetzConnectConfigurator(self.host, self.password)

    # FIXME: for METZ only
    def rs485(self, speed, parity='E'):
        data=0x5300

        parity=parity.upper()
        if parity[0]=='E':
            data|=0x10
        elif parity[0]=='O':
            data|=0x20
        else:
            data|=0x30

        try:
            n=[1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200].index(int(speed))
            data|=(0x1+n)
        except:
            return

        self.checkIdleAfterSend()
        self.client.write_registers(0x41, data, slave=0)
        self.signalMessageTransmission()
        # self.client.diag_restart_communication(False, slave=0)
        # for device in self._devices:
        for device in range(1, 32):
            print("Set speed 0x%02X for device %d" % (data, device))
            self.checkIdleAfterSend()
            self.client.write_registers(0x41, data)
            self.signalMessageTransmission()


class MBIOGatewayMetzConnectConfigurator(object):
    def __init__(self, host, password):
        self._host=host
        self._password=password
        self._link=MBIOSocketString(self._host, 23)
        self._config={}
        self._updated=False
        self._buf=None
        self._auth=False
        self.reset()

    @property
    def config(self):
        return self._config or self

    def reset(self):
        self._buf=''

    def isConnected(self):
        if self._link.isConnected():
            return True

    def isAuthenticated(self):
        if self.isConnected() and self._auth:
            return True
        return False

    def auth(self):
        if self.isConnected():
            if self._auth:
                return True
            passwords=['1234']
            if self._password:
                passwords.insert(0, self._password)
            for password in passwords:
                if self.send(password):
                    self._auth=True
                    self.update()
                    return True
        return False

    def connect(self):
        if self.isConnected():
            return self.auth()
        self._auth=False
        if self._link.connect(timeout=3.0):
            return self.auth()

    def disconnect(self):
        self._link.disconnect()
        self._auth=False
        self.reset()

    def __del__(self):
        self.disconnect()
        self._link=None

    def read(self):
        if self.isConnected():
            data=self._link.read()
            if data:
                self._buf+=data
                return data

    def wait(self, data, timeout=3.0):
        t=time.time()+timeout
        while True:
            r=self.read()
            if r and data in self._buf:
                return self._buf
            if not self._auth and 'password:' in self._buf:
                break
            if not timeout:
                break
            time.sleep(0.01)
            if time.time()>=t:
                break

    def waitprompt(self, timeout=3.0):
        if self.wait('>', timeout):
            try:
                lines=[]
                buf=self._buf.replace('\r\n', '\n')
                for line in buf.split('\n'):
                    line=line.strip()
                    if not line or line=='>':
                        continue
                    lines.append(line)
                return lines
            except:
                pass

    def send(self, data, timeout=3.0):
        if self.isConnected():
            self.reset()
            if self._link.write(data+'\n'):
                return self.waitprompt(timeout)
        return False

    def updateInfo(self):
        if self.isConnected():
            data=self.send('inf')
            if data:
                try:
                    for line in data:
                        pos=line.find(':')
                        if pos>0:
                            key=line[:pos]
                            value=line[pos+1:]
                            key=key.strip().lower()
                            value=value.strip()
                            if key and value:
                                self._config[key]=value
                    return True
                except:
                    pass

    def updateConfig(self):
        if self.isConnected():
            data=self.send('cfg')
            if data:
                try:
                    for line in data:
                        pos=line.find(':')
                        if pos>0:
                            key=line[:pos]
                            value=line[pos+1:]
                            key=key.strip().lower()
                            value=value.strip()
                            if key and value:
                                if 'commands' not in key:
                                    self._config[key]=value
                    return True
                except:
                    pass

    def update(self):
        if self.updateInfo() and self.updateConfig():
            return self._config

    def __getitem__(self, key):
        try:
            return self._config[key.lower()]
        except:
            pass

    def check(self, key, value):
        try:
            if self._config[key.lower()]==value:
                return True
        except:
            pass
        return False

    def command(self, command, value):
        if self.isConnected():
            if command and value is not None:
                command=command.lower()
                value=str(value)
                if value and command in ['name', 'ip', 'mask', 'gateway',
                        'dhcp', 'mac', 'password', 'allowedip', 'allowedmask',
                        'mode', 'parity', 'baudrate', 'stopbits',
                        'timeout', 'terminator', 'idletimeout']:
                    if self.check(command, value):
                        return True
                    self._updated=True
                    self.send('%s %s' % (command, value))
                    if command=='password':
                        self.disconnect()
                    self.update()
                    return self.check(command, value)
        return False

    @property
    def name(self) -> type:
        return self['name']

    @name.setter
    def name(self, value: type):
        self.command('name', value)

    @property
    def password(self) -> type:
        return self['password']

    @password.setter
    def password(self, value: type):
        self.command('password', value)
        self.disconnect()
        self._password=value
        self.connect()

    @property
    def idletimeout(self) -> type:
        return int(self['idletimeout'])

    @idletimeout.setter
    def idletimeout(self, seconds: type):
        seconds=int(seconds)
        if seconds<1:
            seconds=100
        self.command('idletimeout', seconds)

    def setIP(self, ip, mask='255.255.255.0', gw=None):
        try:
            n=ipcalc.Network('%s/%s' % (ip, mask))
            result=self.command('ip', n.to_tuple()[0])
            result&=self.command('mask', str(n.netmask()))
            result&=self.command('dhcp', 0)
            if gw:
                try:
                    n=ipcalc.Network(gw)
                    result&=self.command('gateway', n.to_tuple()[0])
                except:
                    pass
            return result
        except:
            pass
        return False

    def setAllowedIP(self, network):
        try:
            n=ipcalc.Network(network)
            result=self.command('allowedip', n.to_tuple()[0])
            result&=self.command('allowedmask', str(n.netmask()))
            return result
        except:
            pass
        return False

    def setDHCP(self):
        return self.command('dhcp', 1)

    def netrestart(self):
        if self._updated:
            self.send('netstart')
            self._updated=False
            return True

    def setRS485(self, speed=19200, parity='E', stopbits=1, timeout=0.1, terminator=True):
        result=self.command('mode', 0)

        speed=int(speed)
        if speed not in [9600, 19200, 38400, 57600]:
            speed=19200

        result&=self.command('baudrate', speed)
        try:
            parity={'N': 0, 'E': 1, 'O': 2}.get(parity.upper())
            result&=self.command('parity', parity)
        except:
            pass

        if timeout<1:
            timeout=1
        result&=self.command('timeout', int(timeout*1000.0))

        result&=self.command('stopbits', stopbits)
        value=1
        if not terminator:
            value=0
        result&=self.command('terminator', value)
        return result

    def __repr__(self):
        return '%s(%s@%s/%s, fw=%s)' % (self.__class__.__name__, self['name'], self['mac'], self['ip'], self['firmware version'])

    def dump(self):
        t=PrettyTable()
        t.field_names=['Property', 'Value']
        t.align='l'

        for key in self._config:
            t.add_row([key, self._config[key]])

        print(t.get_string())


class MBIOGateways(Items):
    def __init__(self, logger):
        super().__init__(logger)
        self._items: list[MBIOGateway]=[]
        self._itemByHost={}
        self._itemByKey={}
        self._itemByName={}

    def item(self, key):
        item=super().item(key)
        if item:
            return item

        item=self.getByKey(key)
        if item:
            return item

        item=self.getByHost(key)
        if item:
            return item

        item=self.getByName(key)
        if item:
            return item

        try:
            return self[key]
        except:
            pass

    def add(self, item: MBIOGateway) -> MBIOGateway:
        if isinstance(item, MBIOGateway):
            super().add(item)
            self._itemByName[item.name]=item
            self._itemByKey[item.key]=item
            self._itemByHost[item.host]=item

    def getByHost(self, host):
        try:
            return self._itemByHost[host]
        except:
            pass

    def getByName(self, name):
        try:
            return self._itemByName[name]
        except:
            pass

    def getByKey(self, key):
        try:
            return self._itemByKey[key]
        except:
            pass

    def stop(self):
        for item in self._items:
            item.stop()

    def reset(self):
        for item in self._items:
            item.reset()

    def halt(self):
        for item in self._items:
            item.halt()

    def resetHalted(self):
        for item in self._items:
            item.resetHalted()

    def waitForThreadTermination(self):
        for item in self._items:
            item.waitForThreadTermination()

    def discover(self):
        for item in self._items:
            item.discover()

    # def dump(self):
        # if not self.isEmpty():
            # t=PrettyTable()
            # t.field_names=['#', 'Key', 'Name', 'Host', 'Open']
            # t.align='l'
            # for item in self._items:
                # t.add_row([self.index(item), item.key, item.name, item.host, item.isOpen()])

        # print(t.get_string(sortby="Key"))


if __name__ == "__main__":
    pass



#!/bin/python

from .device import MBIODevice
from .xmlconfig import XMLConfig


class MBIODeviceEBM(MBIODevice):
    def onInit(self):
        self._vendor='EBM'
        self._model='BASE'

        # must return FAN address
        self.setPingHoldingRegister(0xd100)

        self.config.set('enablesource', 'bus')
        self.config.set('speedsource', 'bus')
        self.config.set('controlmode', 'speed')
        self.config.set('maxrpm', 0)
        self.config.set('ramp')
        self.config.set('ramp0')
        self.config.set('ramp1')
        self.config.set('io1type', 'di')
        self.config.set('io2type', 'ai')
        self.config.set('io3type', 'ao')
        self.config.set('watchdogsource')
        self.config.set('vref', 0)
        self.config.set('iref', 0)

        self.value('speed', unit='%')

        self.value('rpm', unit='t/min', resolution=1.0)
        self.value('status', unit=35)
        self.value('warning', unit=35)
        self.value('t1', unit='C')
        self.value('t2', unit='C')
        self.value('t3', unit='C')
        self.valueDigital('clockwise')
        self.value('dci', 'A', resolution=0.2)
        self.value('pow', 'W', resolution=1.0)
        self.value('ene', 'kWh', resolution=0.001)
        self.value('hcnt', unit='h')

        self._timeoutRefreshSlow=0

    def sendWarmReset(self):
        """Reset motor (ack ERROR + reload CONFIG) without communication break"""
        return self.writeRegisters(0xd000, 1)

    def sendColdreset(self):
        """Reset motor (ack ERROR + reload CONFIG) with communication break"""
        return self.writeRegisters(0xd000, 8)

    def onLoad(self, xml: XMLConfig):
        item=xml.child('enable')
        if item:
            self.config.update('enablesource', item.get('enablesource'))
            if self.config.enablesource=='bus':
                self.valueDigital('enable', writable=True)
            else:
                self.valueDigital('enable')

        subitem=xml.child('io1')
        if subitem:
            self.config.update('io1type', subitem.get('type'))

        subitem=xml.child('io2')
        if subitem:
            self.config.update('io2type', subitem.get('type'))

        subitem=xml.child('io3')
        if subitem:
            self.config.update('io3type', subitem.get('type'))

        item=xml.child('speed')
        if item:
            self.config.update('speedsource', item.get('source'))
            if self.config.speedsource=='bus':
                value=self.value('sp', unit='%', writable=True, resolution=0.1)
                value.setRange(0, 100)

            self.config.set('x0', 1)
            self.config.set('y0', 0)
            self.config.set('x1', 10)
            self.config.set('y1', 100)

            subitem=item.child('input')
            if subitem:
                self.config.update('x0', subitem.getFloat('x0', vmin=0, vmax=10))
                self.config.update('x1', subitem.getFloat('x1', vmin=self.config.x0, vmax=10))
                self.config.update('y0', subitem.getFloat('y0', vmin=0, vmax=100))
                self.config.update('y1', subitem.getFloat('y1', vmin=self.config.y0, vmax=100))

            subitem=item.child('sensor')
            if subitem:
                self.config.set('sensorsource', 'ai1')
                self.config.update('x0', subitem.getFloat('x0', vmin=0, vmax=10))
                self.config.update('x1', subitem.getFloat('x1', vmin=self.config.x0, vmax=10))
                self.config.update('y0', subitem.getFloat('y0', vmin=0, vmax=100))
                self.config.update('y1', subitem.getFloat('y1', vmin=self.config.y0, vmax=100))
                self.value('ai1', 'V')
                self.value('ai2', 'V')

            subitem=item.child('watchdog')
            if subitem:
                self.config.set('watchdogsource', subitem.get('source'))
                self.config.set('watchdogsp', subitem.getInt('sp', 0, vmin=0))
                if self.config.watchdogsource=='bus':
                    self.config.set('watchdogdelay', subitem.getInt('delay', 60, vmin=0))
                else:
                    self.config.set('watchdoglevel', subitem.getInt('min', 0, vmin=0))

            self.config.update('controlmode', item.get('control'))
            self.config.update('ramp', item.getInt('ramp', vmin=0))
            self.config.update('ramp0', item.getInt('ramp0', vmin=0))
            self.config.update('ramp1', item.getInt('ramp1', vmin=0))

    def poweron(self):
        # paramerter set 0
        self.writeRegistersIfChanged(0xd105, 0)

        # reload vref, iref
        r=self.readHoldingRegisters(0xd1a0, 2)
        if r:
            self.config.vref=r[0]*20.0/1000.0  # V
            self.config.iref=r[1]*2/1000.0     # A

        # enable source
        # v6+
        data={'none': 0, 'bus': 1, 'di1-off': 2, 'di2-off': 3, 'di3-off': 4, 'di1-on': 5, 'di1': 5, 'di2-on': 6, 'di2': 6}
        self.writeRegistersIfChanged(0xd16a, data.get(self.config.enablesource, 0))

        # speed source
        self.writeRegistersIfChanged(0xd16c, 0)

        # control mode
        data={'speed': 0, 'sensor': 1}
        self.writeRegistersIfChanged(0xd106, data.get(self.config.controlmode, 0))

        # start/min/max modulation levels
        # TODO: not modifiable ?
        # self.writeRegistersIfChanged(0xd110, 0)
        # self.writeRegistersIfChanged(0xd10e, 255)

        # stop enable
        self.writeRegistersIfChanged(0xd112, 1)

        # max speed
        r=self.readHoldingRegisters(0xd119, 1)
        if r:
            self.config.update('maxrpm', r[0])

        ramp=self.config.ramp1 or self.config.ramp
        if ramp is not None:
            self.writeRegistersIfChanged(0xd11f, int(float(ramp)/2.5))
        ramp=self.config.ramp0 or self.config.ramp
        if ramp is not None:
            self.writeRegistersIfChanged(0xd120, int(float(ramp)/2.5))

        # input
        if self.config.speedsource != 'bus':
            self.writeRegistersIfChanged(0xd12a, int(self.config.x0/10.0*65535.0))

            self.writeRegistersIfChanged(0xd12c, int(self.config.x1/10.0*65535.0))
            if self.config.controlmode=='speed':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*64000.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*64000.0))
            elif self.config.controlmode=='sensor':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*65535.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*65535.0))
            elif self.config.controlmode=='openloop':
                self.writeRegistersIfChanged(0xd12b, int(self.config.y0/100.0*65535.0))
                self.writeRegistersIfChanged(0xd12d, int(self.config.y1/100.0*65535.0))

        # output speed monitoring 0-10V
        # self.writeRegistersIfChanged(0xd130, 0)
        self.writeRegistersIfChanged(0xd140, 0)
        self.writeRegistersIfChanged(0xd141, 0)
        self.writeRegistersIfChanged(0xd142, 64000)
        self.writeRegistersIfChanged(0xd143, 64000)

        # speed source
        data={'bus': 1, 'ai1': 0, 'ai2': 2}
        self.writeRegistersIfChanged(0xd101, data.get(self.config.speedsource, 1))

        if self.config.controlmode=='sensor':
            data={'ai1': 0, 'ai2': 1, 'max': 2, 'min': 3, 'mean': 4}
            self.writeRegistersIfChanged(0xd147, data.get(self.config.sensorsource, 0))

        # shedding
        self.writeRegistersIfChanged(0xd150, 0)

        # IO1
        data={'di': 0, 'ai': 2}
        self.writeRegistersIfChanged(0xd158, data.get(self.config.io1type, 0))

        # IO2
        data={'di': 0, 'ai': 2}
        self.writeRegistersIfChanged(0xd159, data.get(self.config.io1type, 2))

        # IO3
        data={'di': 0, 'di0': 1, 'di!': 1, 'ao': 3}
        self.writeRegistersIfChanged(0xd15a, data.get(self.config.io1type, 3))

        # watchdog
        if self.config.watchdogsource:
            data={'bus': 1, 'ai1': 2, 'ai2': 2}
            self.writeRegistersIfChanged(0xd15c, data.get(self.config.watchdogsource, 0))
            if self.config.watchdogsource=='bus':
                self.writeRegistersIfChanged(0xd15e,
                    int(self.config.getInt('watchdogdelay', default=15, vmin=10)/100.0))
            elif self.config.watchdogsource in ['ai1', 'ai2']:
                self.writeRegistersIfChanged(0xd15f,
                    self.config.getFloat('watchdoglevel', default=0, vmin=0)/10.0*65535.0)

            v=self.config.getInt('value', default=0, vmin=0, vmax=100)
            if self.config.controlmode=='speed':
                self.writeRegistersIfChanged(0xd15d, int(v/100.0*64000.0))
            else:
                self.writeRegistersIfChanged(0xd15d, int(v/100.0*65535.0))

            # keep rotation direction
            self.writeRegistersIfChanged(0xd15b, 2)
        else:
            self.writeRegistersIfChanged(0xd15c, 0)

        if self.config.controlmode=='sensor':
            # Sensor caracteristics
            encoder=self.encoder()
            encoder.float32(0)
            encoder.float32(100)
            encoder.writeRegistersIfChanged(0xd160)

        return True

    def poweronsave(self):
        self.sendWarmReset()

    def poweroff(self):
        value=self.values.enable
        if value is not None:
            value.off()
        value=self.values.sp
        if value is not None:
            value.set(0)

        self.sync()
        return True

    def refresh(self):
        if self.config.maxrpm is not None:
            r=self.readInputRegisters(0xd010, 1)
            if r:
                try:
                    self.values.rpm.updateValue(r[0]/64000.0*self.config.maxrpm)
                except:
                    pass

        # Max 9 registers with firwmare <6
        r=self.readInputRegisters(0xd011, 9)
        if r:
            self.values.status.updateValue(r[0])
            self.values.warning.updateValue(r[1])
            # self.values.dcv.updateValue(r[2]/255.0*self.config.vref)
            self.values.dci.updateValue(r[3]/255.0*self.config.iref)
            self.values.t1.updateValue(r[4])
            self.values.t2.updateValue(r[5])
            self.values.t3.updateValue(r[6])
            self.values.speed.updateValue(r[8]/65535.0*100.0)

        if self.isTimeout(self._timeoutRefreshSlow):
            r=self.readInputRegisters(0xd023, 8)
            if r:
                if self.values.ai1 is not None:
                    self.values.ai1.updateValue(r[0]/65535.0*10.0)
                if self.values.ai2 is not None:
                    self.values.ai2.updateValue(r[0]/65535.0*10.0)
                decoder=self.decoderFromRegisters(r[6:])
                self.values.ene.updateValue(decoder.dword())

            r=self.readInputRegisters(0xd021, 1)
            if r:
                v=r[0]/65535.0*self.config.vref*self.config.iref
                self.values.pow.updateValue(v)

            r=self.readHoldingRegisters(0xd009, 1)
            if r:
                self.values.hcnt.updateValue(r[0])

            self._timeoutRefreshSlow=self.timeout(15)

        return 3.0

    def sync(self):
        value=self.values.enable
        if value is not None and value.isPendingSync():
            if self.writeRegisters(0xd00f, value.toReachValue):
                value.clearSyncAndUpdateValue()

        value=self.values.sp
        if value is not None and value.isPendingSync():
            vref=65535.0
            if self.config.controlmode=='speed':
                vref=64000.0
            if self.writeRegisters(0xd001, int(value.toReachValue/100.0*vref)):
                value.clearSyncAndUpdateValue()


if __name__ == "__main__":
    pass
